"""   HEEKSFUNCTIONS
6.12.2017 by jean-claude.feltes@education.lu
This is a collection of functions to draw a HeeksCAD 2D drawing
and add machining operations like drilling, pocketing and profiling


HOW TO USE HEEKSFUNCTIONS

1. openheeks()
   eventually change default settings with
   set_feedrate, set_clearanceheight
    
2. draw with  
    point(x,y)
    line (x1, y1, x2, y2)
    circle(x,y,r)
    rect(x, y, length, height)
    roundedrect( x, y, length, height, r, *name)
    arc (xcenter,ycenter, xbegin, ybegin, xend, yend)     
    arc_ang(xcenter, ycenter, r, alphabegin, alphaend)
    arc_e(xcenter, ycenter, r, excentricity, *name)
    
    
    add operations like 
    profile, pocket, drilling immediately after each sketch (circle, rect ...)
    
    eventually add holding tags with     add_tag 
    immediately before calling profile
    
3. closeheeks()
4. showheeks() to start HeeksCAD and show the drawing    


In HeeksCAD, use Menu Machining - Postprocess to generate the G code
    
------------------------------------------------------------------
Known bugs:
There is still a problem with tools:
HeeksCAD behaves differently if 
* it is started without a file, then Open file -> correct function, tooltable from XML is loaded
* or started via command line with the file to load in command line -> default tools are loaded
Workaround: at first use load 12345.tooltable for tools, and set as default tools.
From then on it should work.

Tools are numbered according to their diameter.
The tooltable 12345.tooltable contails mills from 1 to 5 mm diameter

    
"""
import os
from math import sin, cos, asin,  pi

#-------------------------------------------------------------------------------------------
""" default settings for CNC operations """

# feed rates
xyfeed=800
zfeed=150

# pocket, profile and drilling
clear_height = 20.0
spin=7000

# pocket + profile
pattern=0
surface=0

#drilling
zrapid = 5.0

# pocket
cutmode = 0
from_center=1
use_zig_zag=0
zig_angle = 90

# tags
tag_width = 1
tag_angle = 80
tag_height = 2
tag_string = ""

#----------------------------------------------------------------------
# predefined constants for profile operation

INSIDE = -1
OUTSIDE = 1
ON_LINE = 0

#---------------------------------------------------------------------

# variables to remember coordinates (for special operations)
x_center=0
y_center=0
xbegin=0
ybegin=0
xend=0
yend=0


#-------------------------------------------------------------------------------------------------
# XML functions for Heeks

# global
ID = 0
last_sketch_ID = 0
heeksstring = ""
drawing=""
tmpfile="tmp_10_heeks.heeks"

# Heeks XML
preamble="""<?xml version="1.0" encoding="UTF-8" ?>
<HeeksCAD_Document>
    <Program machine="LinuxCNC" output_file="/tmp/heekscnc-L7htaV.ngc" output_file_name_follows_data_file_name="1" program="" units="1" ProgramPathControlMode="3" ProgramMotionBlendingTolerance="0.0001" ProgramNaiveCamTolerance="0.0001" title_from_id="1" id="1">
"""        
        
        
endprogram="""        
    <nccode edited="0" id="1" />
    </Program>
"""    


tools="""
<Tools>
    <Tool title="1 mm End Mill" tool_number="1" id="1">
        <params diameter="1" tool_length_offset="0" automatically_generate_title="1" material="1" type="end_mill" corner_radius="0" flat_radius="0" cutting_edge_angle="0" cutting_edge_height="50.8" />
    </Tool>
    <Tool title="2 mm End Mill" tool_number="2" id="2">
        <params diameter="2" tool_length_offset="0" automatically_generate_title="1" material="1" type="end_mill" corner_radius="0" flat_radius="0" cutting_edge_angle="0" cutting_edge_height="50.8" />
    </Tool>
    <Tool title="3 mm End Mill" tool_number="3" id="3">
        <params diameter="3" tool_length_offset="0" automatically_generate_title="1" material="1" type="end_mill" corner_radius="0" flat_radius="0" cutting_edge_angle="0" cutting_edge_height="50.8" />
    </Tool>
    <Tool title="4 mm End Mill" tool_number="4" id="4">
        <params diameter="4" tool_length_offset="0" automatically_generate_title="1" material="1" type="end_mill" corner_radius="0" flat_radius="0" cutting_edge_angle="0" cutting_edge_height="50.8" />
    </Tool>
    <Tool title="5 mm End Mill" tool_number="5" id="5">
        <params diameter="5" tool_length_offset="0" automatically_generate_title="1" material="1" type="end_mill" corner_radius="0" flat_radius="0" cutting_edge_angle="0" cutting_edge_height="50.8" />
    </Tool>
</Tools>    
<Patterns />
<Surfaces />
<Stocks />
        """

operations="<Operations>\n"
endoperations="</Operations>\n"
ending="</HeeksCAD_Document>\n"

#--------------------------------------------------------------------------------------------
# functions for begin and end of heeks file
def openheeks():
    """put XML preamble at the beginning of heeksstring"""
    global heeksstring, tools
    
    heeksstring = preamble + tools

def closeheeks():
    """put XML ending at the end of heeksstring and print XML""" 
    global heeksstring, operations, endprogram, drawing,  ending 
    heeksstring += operations 
    heeksstring += endoperations 
    heeksstring += endprogram 
    heeksstring += drawing
    heeksstring += ending  
    
    print heeksstring     

def showheeks():    
    """write heeksstring to temporary file and start heekscad to show the drawing"""
    f= open(tmpfile, 'w')
    f.write(heeksstring)
    f.close()
    
    os.system("heekscad "+ tmpfile)

#----------------------------------------------------------------------------------------------
# 2D drawing functions

def opensketch(*name):
    global ID, last_sketch_ID, drawing
    ID +=1
    last_sketch_ID = ID
    if not len(name):
        s = '<Sketch title="Sketch" title_from_id="1" id="%d">\n' %(ID)
    else:
        s = '<Sketch title="%s" title_from_id="0" id="%d">\n' %(name[0], ID)
    drawing+= s
        
def closesketch():
    global  drawing
    s = '</Sketch>\n'
    drawing+= s
       
            
def point(x,y, *name):
    """append point to XML drawing
    name is optional"""
    
    global ID, last_sketch_ID, drawing
    ID+=1
    last_sketch_ID = ID
    
    if not len(name):
        s  = '<Point col="0" x="%g" y="%g" z="0" title_from_id="1" id="%d" />' %(x,y,ID) 
    else:
        s = '<Point col="0" x="%g" y="%g" z="0" title_from_id="0" title="%s" id="%d" />' %(x,y,name[0],ID)  
    
    drawing+= "\t\t" + s + "\n"

def line(x1, y1, x2, y2, *name):
    """append line to XML drawing
    from (x1,y1) to (x2,y2)
    name is optional, if no name, name is generated from ID
    """
    global ID, drawing
    ID +=1
    if not len(name):
        # line with no name
        s = '\t<Line col="0" sx="%g" sy="%g" sz="0" ex="%g" ey="%g" ez="0" title_from_id="1" id="%d" />' %(x1,y1,x2,y2,ID) 
    else:
        # Heeks Line has no title, so we must use a sketch if line should be named
        s1= '<Sketch title="%s" title_from_id="0" id="%d">\n' %(name[0], ID)
        ID+=1 
        s = '\t<Line col="0" sx="%g" sy="%g" sz="0" ex="%g" ey="%g" ez="0" title_from_id="0"  id="%d" />\n' %(x1,y1,x2,y2,ID) 
        s2= '</Sketch>\n'
        s="\t" + s1 +"\t" + s + "\t"+ s2
    drawing+= s+ "\n" 


def circle(x,y,r, *name):
    """append circle to XML drawing
    x, y = center,  r = radius
    name is optional, if no name, name is generated from ID
    """
    global ID, last_sketch_ID, drawing
    ID +=1
    last_sketch_ID = ID 
    if not len(name):
        s = '\t<Circle col="0" r="%g" cx="%g" cy="%g" cz="0" ax="-0" ay="-0" az="-1" title_from_id="1" id="%d" />' %(r,x,y,ID) 
    else:
        s = '\t<Circle col="0" r="%g" cx="%g" cy="%g" cz="0" ax="-0" ay="-0" az="-1" title_from_id="0" title="%s" id="%d" />' %(r,x,y,name[0], ID) 
    
    drawing+= "\t\t" + s + "\n"
    

def rect(x, y, length, height, *name):
    """append rectangle to XML drawing
    x, y = center,  r = radius
    name is optional, if no name, name is generated from ID
    """
    global ID, last_sketch_ID, drawing
    ID +=1
    last_sketch_ID = ID
    if not len(name):
        s0 = '<Sketch title="RECTANGLE" title_from_id="0" id="%d">\n' %(ID)
    else:
        s0 = '<Sketch title="%s" title_from_id="0" id="%d">\n' %(name[0], ID)
            
    ID+=1
    s1 = '\t<Line col="0" sx="%g" sy="%g" sz="0" ex="%g" ey="%g" ez="0" title_from_id="1" id="%d" />\n' %(x,y,x+length,y,ID) 
    ID+=1
    s2 = '\t<Line col="0" sx="%g" sy="%g" sz="0" ex="%g" ey="%g" ez="0" title_from_id="1" id="%d" />\n' %(x+length,y,x+length,y+height,ID) 
    ID+=1
    s3 = '\t<Line col="0" sx="%g" sy="%g" sz="0" ex="%g" ey="%g" ez="0" title_from_id="1" id="%d" />\n' %(x+length,y+height,x,y+height,ID)
    ID+=1
    s4 = '\t<Line col="0" sx="%g" sy="%g" sz="0" ex="%g" ey="%g" ez="0" title_from_id="1" id="%d" />\n' %(x,y+height,x,y,ID)
    s5 = '</Sketch>\n'
    drawing+= s0+s1+s2+s3+s4+s5    

def arc_ang(xcenter, ycenter, r, alphabegin, alphaend):
    """ append unnamed arc to drawing"""
    global x_center, y_center, xbegin, ybegin, xend, yend           # to remember these coordinates
    x_center=xcenter
    y_center = ycenter
    xbegin = r*cos(alphabegin * pi/180.0) + xcenter
    ybegin = r*sin(alphabegin * pi/180.0) + ycenter
    xend = r*cos(alphaend * pi/180.0) + xcenter
    yend = r*sin(alphaend * pi/180.0) +ycenter 
    arc(xcenter,ycenter, xbegin, ybegin, xend, yend )

def arc(xcenter,ycenter, xbegin, ybegin, xend, yend ):
    """ append arc without  name to drawing"""
    global ID, drawing
    ID +=1
    s = '\t<Arc cx="%g" cy="%g" cz="0" ax="0" ay="0" az="1" col="0" sx="%g" sy="%g" sz="0" ex="%g" ey="%g" ez="0" id="%d" />\n' %(xcenter,ycenter, xbegin, ybegin, xend, yend, ID)
    drawing+= s+ "\n" 

def arc_e(xcenter, ycenter, r, excentricity, *name):
    """ append excentric 'half'-circle in horizontal position
    with horizontal line
    for motor axe
    excentricity is the vertical distance to the horizontal line"""
    global x_center, y_center, xbegin, ybegin, xend, yend           # to remember these coordinates
    x_center=xcenter
    y_center = ycenter
    opensketch(*name)
    alpha = asin(excentricity*1.0/r)            # *1.0 forces float
    print alpha
    xbegin = r*cos(pi-alpha) + xcenter
    ybegin = r*sin(pi-alpha) + ycenter
    xend = r*cos(alpha) + xcenter
    yend = r*sin(alpha) +ycenter 
    arc(xcenter,ycenter, xbegin, ybegin, xend, yend )
    line(xbegin, ybegin, xend, yend)
    closesketch()

def get_xy_begin_end():
    """ returns begin and end of last arc"""
    global xbegin, ybegin, xend, yend     
    return xbegin, ybegin, xend, yend

def get_center():
    """gets center of last arc"""
    global x_center, y_center
    return  x_center, y_center  
        
def roundedrect( x, y, length, height, r, *name):
    """append rectangle with rounded vertices to drawing"""
    opensketch(*name)
    line(x+r, y, x+length-r, y)
    arc(x+length-r, y+r, x+length-r, y, x+length, y+r)
    line(x+length, y+r, x+length, y+height-r)
    arc(x+length-r, y+height-r, x+length, y+height-r, x+length-r, y+height)
    line(x+length-r, y+height, x+r, y+height)
    arc(x+r, y+height-r, x+r, y+height, x, y+height-r)
    line(x, y+height-r, x, y+r)
    arc(x+r, y+r, x, y+r, x+r, y)
    closesketch()
       
# ----------------------------------------------------------------------------------------
#    Operations
def set_feedrate(xy, z):
    global xyfeed, zfeed
    xyfeed=xy
    zfeed=z
    
def set_clearanceheight(h):
    global clear_height
    clear_height = h


def pocket(toolnumber, startdepth, depth, stepdown, stepover):
    """adds a pocket operation to the last drawn sketch"""
    ##toolnumber=4        # TODO: this sets diameter implicitely! Must be done explicitely
   
      
    global operations, ID, last_sketch_ID
    ID +=1
    s="""<Pocket sketch="%d" active="1" tool_number="%d" pattern="%d" surface="%d" title_from_id="1" id="%d">
                <params step="%g" cut_mode="%d" mat="0" from_center="%d" keep_tool_down="1" use_zig_zag="%d" zig_angle="%d" zig_unidirectional="0" entry_move="0" />
                <depthop clear="%g" down="%g" userdepths="" startdepth="%g" depth="%g" r="10" />
                <speedop hfeed="%g" vfeed="%g" spin="%g" />
    </Pocket>
    """  %(last_sketch_ID, toolnumber, pattern, surface, ID, stepover, cutmode,from_center, use_zig_zag, zig_angle, clear_height, stepdown, startdepth, depth, xyfeed, zfeed, spin)
    
    operations += s
    

# ...........................................................................................................

def profile(toolnumber, side, startdepth, depth, stepdown):
    """side:  1 = outside, -1 = inside, 0 = on line
    tags can be added with add_tag, but add_tag must be called BEFORE profile!
    """
    
    ##toolnumber=1        # TODO: this sets diameter implicitely! Must be done explicitely
    ##side=0           # 1 = outside, -1 = inside, 0 = on
    
    
    
    spin=7000
    
    global operations, ID, last_sketch_ID, tag_string 
    ID +=1
    
    s="""<Profile sketch="%d" active="1" tool_number="%d" pattern="%d" surface="%d" title_from_id="1" id="%d">
                <params side="%d" cut_mode="0" auto_roll_on="1" auto_roll_off="1" roll_radius="2" start_given="0" end_given="0" sort_sketches="1" extend_at_start="0" extend_at_end="0" lead_in_line_len="0" lead_out_line_len="0" offset_extra="0" do_finishing_pass="0" only_finishing_pass="1" finishing_feed_rate="100" finish_cut_mode="0" finishing_step_down="1" />
                <depthop clear="%g" down="%g" userdepths="" startdepth="%g" depth="%g" r="10" />
                <speedop hfeed="%g" vfeed="%g" spin="%g" />
                <Tags>
                "%s"
                </Tags>
        </Profile>
    """ %(last_sketch_ID, toolnumber, pattern, surface, ID, side, clear_height, stepdown, startdepth, depth, xyfeed, zfeed, spin, tag_string )
    
    operations += s 
    
    tag_string=""   
#..........................................................................................

def add_tag(x,y):
    """ adds tags to a profile
    add_tag must be used BEFORE the profile function"""
    global tag_string    
    tag_string +="""<Tag x="%g" y="%g" width="%g" angle="%g" height="%g" />
    """%(x, y, tag_width, tag_angle, tag_height)
    

    
#...............................................................................................

def drilling(tool_number, startdepth, depth, stepdown):

    global operations, ID, last_sketch_ID
    ID +=1
    
    s="""<Drilling active="1" tool_number="%d" pattern="0" surface="0" title_from_id="1" id="%d">
                <params dwell="0" retract_mode="0" spindle_mode="0" internal_coolant_on="0" rapid_to_clearance="1" />
                <Point id="%d" />
                <depthop clear="%g" down="%g" userdepths="" startdepth="%g" depth="%g" r="%g" />
                <speedop hfeed="800" vfeed="150" spin="7000" />
        </Drilling>
    """ %(tool_number, ID, last_sketch_ID, clear_height, stepdown, startdepth, depth, zrapid)
    
    operations += s    
#..............................................................................................
    

#_________________________________________________________________________________________
# TEST PROGRAM
""" For all functions, the name argument is optional.
If not needed, it can be omitted."""
        
if __name__ == '__main__':
    
    set_feedrate(500,100)           # set_feedrate(xy, z)
    set_clearanceheight(15)
    
    # open new drawing:
    openheeks()
    
    # draw 2D objects with or without machining operation 
    
    point (0,10, "P01")
    drilling(3, 0, -6, 2)           # drilling(tool_number, startdepth, depth, stepdown)
    
    line(0,0,20,20, "diagonal")     # line(x1, y1, x2, y2, *name)
    # unfortunately this does not work:
    profile(1, ON_LINE, 0, -1, 1)  # profile(toolnumber, side, startdepth, depth, stepdown)
    
    circle(20,10,6, "test-c")       # circle(x,y,radius, *name)
    pocket(5, 0,-5,3,2)             # pocket(toolnumber, startdepth, depth, stepdown, stepover)
    
    
    rect(40,-10, 10, 25, "Plate")
    add_tag(0, 15)
    add_tag(30, 20)
    profile(2, INSIDE, 0, -3, 0.5)   # profile(toolnumber, side, startdepth, depth, stepdown)
    
    
    roundedrect(-30,-30, 25,15,5, "RoundedRect")
    add_tag(-20, -15)
    add_tag(-20, -30)
    profile(2, OUTSIDE, 0, -3, 1) 
    
    arc(30,30,30,20,30,30)          # arc(xcenter,ycenter, xbegin, ybegin, xend, yend )
    arc_ang(-20,0,20, +60, 90)     # arc_ang(xcenter, ycenter, r, alphabegin, alphaend)
    arc_e(-20, 0, 10, 2)            # arc_e(xcenter, ycenter, r, excentricity, *name)
    
    # helper functions for arc can be used to drill at extreme points and center:
    xb, yb, xe, ye = get_xy_begin_end()
    point(xb,yb)
    point(xe,ye) 
    x,y = get_center()  
    point(x,y)
    
    # It is easy to draw multiple objects with a for loop:
    for i in range(0,5):
        circle (i*10,-20,1.5)
        point (i*10,-20, "P"+str(i)) 
    
    # complex objects can consist of multiple lines
    opensketch("triangle")
    line (0,-5,30,-5)
    line(30,-5,30,5)
    line(30,5,0,-5)
    closesketch()
    pocket(1, 0,-5,3,2) 
        
    closeheeks()
    showheeks()
